use crate::{asset_loader::AssetLoadState::AssetsLoaded, system_set::MovementSet};
use bevy::{prelude::*, sprite::Anchor, window::PrimaryWindow};

pub struct BackgroundPlugin;

impl Plugin for BackgroundPlugin {
    fn build(&self, app: &mut App) {
        app.add_systems(OnEnter(AssetsLoaded), setup).add_systems(
            Update,
            update_parallax.in_set(MovementSet::PostCameraMovement),
        );
    }
}

#[derive(Component)]
struct Parallax {
    speed: f32,
    tile_width: u32,
    start_offset: f32,
}

fn setup(
    mut commands: Commands,
    asset_server: Res<AssetServer>,
    assets: Res<Assets<Image>>,
    window: Query<&Window, With<PrimaryWindow>>,
) {
    let window = window.single();

    let ground = asset_server.load("textures/groundSnow.png");
    let ground_size = assets.get(&ground).unwrap().size();
    let ground_tiles_needed = (window.width() * 4.0) as u32 / ground_size.x;
    let ground_tile_start_offset = -1.5 * window.width();
    commands.spawn((
        SpriteBundle {
            texture: ground,
            sprite: Sprite {
                custom_size: Some(Vec2::new(
                    (ground_size.x * ground_tiles_needed) as f32,
                    ground_size.y as f32,
                )),
                anchor: Anchor::BottomLeft,
                ..default()
            },
            transform: Transform::from_xyz(ground_tile_start_offset, -0.5 * window.height(), 10.0),
            ..default()
        },
        ImageScaleMode::Tiled {
            tile_x: true,
            tile_y: true,
            stretch_value: 1.0,
        },
        Parallax {
            speed: 0.0, // same speed as the camera
            tile_width: ground_size.x,
            start_offset: ground_tile_start_offset,
        },
    ));

    let background = asset_server.load("textures/background.png");
    let background_size = assets.get(&background).unwrap().size();
    let background_tiles_needed = (window.width() * 4.0) as u32 / background_size.x;
    let background_tile_start_offset = -1.5 * window.width();
    commands.spawn((
        SpriteBundle {
            texture: background,
            sprite: Sprite {
                custom_size: Some(Vec2::new(
                    (background_size.x * background_tiles_needed) as f32,
                    background_size.y as f32,
                )),
                anchor: Anchor::BottomLeft,
                ..default()
            },
            transform: Transform::from_xyz(
                background_tile_start_offset,
                -0.5 * window.height(),
                -1.0,
            ),
            ..default()
        },
        ImageScaleMode::Tiled {
            tile_x: true,
            tile_y: true,
            stretch_value: 1.0,
        },
        Parallax {
            speed: 0.5, // same speed as the camera
            tile_width: background_size.x,
            start_offset: background_tile_start_offset,
        },
    ));
}

fn update_parallax(
    mut parallax: Query<(&Parallax, &mut Transform), Without<Camera2d>>,
    camera: Query<&Transform, With<Camera2d>>,
) {
    let Ok(camera) = camera.get_single() else {
        return;
    };

    for (parallax, mut transform) in &mut parallax {
        // + parallax.start_offset
        let mut new_offset = camera.translation.x * parallax.speed;
        let tiling_undershoot =
            ((camera.translation.x - new_offset) / parallax.tile_width as f32) as i32;
        if tiling_undershoot != 0 {
            new_offset += parallax.tile_width as f32 * tiling_undershoot as f32;
        }
        transform.translation.x = new_offset + parallax.start_offset;
    }
}
