use crate::{
    asset_loader::AssetLoadState::AssetsLoaded,
    player::Player,
    system_set::{DespawnSet, MovementSet},
};
use std::time::Duration;

use bevy::{prelude::*, sprite::Anchor, window::PrimaryWindow};
use rand::RngCore;

const OBSTACLE_SPAWN_DELAY: f32 = 2.0;

pub struct ObstaclePlugin;

impl Plugin for ObstaclePlugin {
    fn build(&self, app: &mut App) {
        app.add_systems(OnEnter(AssetsLoaded), setup)
            .add_systems(
                Update,
                add_obstacles.in_set(MovementSet::PostPlayerMovement),
            )
            .add_systems(
                Update,
                despawn_offscreen.in_set(DespawnSet::DespawnOffscreen),
            );
    }
}

#[derive(Component)]
struct ObstacleTimer {
    timer: Timer,
    obstacles_placed: u32,
}

#[derive(Component)]
pub struct Obstacle {
    pub p1: Vec2,
    pub p2: Vec2,
    pub p3: Vec2,
}

#[derive(Component)]
pub struct Star {
    pub radius: f32,
    pub points: u32,
}

fn setup(mut commands: Commands) {
    commands.spawn(ObstacleTimer {
        timer: Timer::new(
            Duration::from_secs_f32(OBSTACLE_SPAWN_DELAY),
            TimerMode::Repeating,
        ),
        obstacles_placed: 0,
    });
}

fn add_obstacles(
    mut commands: Commands,
    mut timer: Query<&mut ObstacleTimer>,
    window: Query<&Window, With<PrimaryWindow>>,
    player: Query<&Transform, With<Player>>,
    asset_server: Res<AssetServer>,
    assets: Res<Assets<Image>>,
    time: Res<Time>,
) {
    let window = window.single();

    let Ok(mut timer) = timer.get_single_mut() else {
        return;
    };

    let Ok(player) = player.get_single() else {
        return;
    };

    timer.timer.tick(time.delta());

    if timer.timer.just_finished() {
        let mut rng = rand::thread_rng();

        let is_roof_obstacle = rng.next_u32() % 2 == 0;
        let obstacle_scale = if rng.next_u32() % 2 == 0 { 1.0 } else { 0.75 };

        let texture: Handle<Image> = match rng.next_u32() % 4 {
            0 => asset_server.load("textures/rock.png"),
            1 => asset_server.load("textures/rockGrass.png"),
            2 => asset_server.load("textures/rockIce.png"),
            3 => asset_server.load("textures/rockSnow.png"),
            _ => unreachable!(),
        };

        let transform = Transform::from_xyz(
            window.width() + player.translation.x,
            (if is_roof_obstacle { 1.0 } else { -1.0 }) * window.height() / 2.0,
            5.0,
        ) * Transform::from_scale(Vec3::splat(obstacle_scale));

        let obstacle_texture = assets.get(&texture).unwrap();
        let obstacle_width = (obstacle_texture.width() as f32 * obstacle_scale) as f32;
        let obstacle_height = (obstacle_texture.height() as f32 * obstacle_scale) as f32;

        let obstacle_tip = Vec2::new(
            transform.translation.x,
            transform.translation.y
                + (if is_roof_obstacle {
                    -1.0 * obstacle_height
                } else {
                    obstacle_height
                }),
        );
        let obstacle_left = Vec2::new(
            transform.translation.x - obstacle_width / 2.0,
            transform.translation.y,
        );
        let obstacle_right = Vec2::new(
            transform.translation.x + obstacle_width / 2.0,
            transform.translation.y,
        );

        commands.spawn((
            SpriteBundle {
                sprite: Sprite {
                    anchor: if is_roof_obstacle {
                        Anchor::TopCenter
                    } else {
                        Anchor::BottomCenter
                    },
                    flip_y: is_roof_obstacle,
                    ..default()
                },
                texture: texture,
                transform: transform,
                ..default()
            },
            Obstacle {
                p1: obstacle_tip,
                p2: obstacle_left,
                p3: obstacle_right,
            },
        ));

        timer.obstacles_placed += 1;

        // Place star at tip of obstacle
        let star_value = rng.next_u32() % 4;
        if star_value != 0 {
            let texture: Handle<Image> = match star_value {
                1 => asset_server.load("textures/starBronze.png"),
                2 => asset_server.load("textures/starSilver.png"),
                3 => asset_server.load("textures/starGold.png"),
                _ => unreachable!(),
            };

            let star_texture = assets.get(&texture).unwrap();
            let star_width = star_texture.width() as f32;

            let transform = Transform::from_xyz(
                transform.translation.x + star_width / 4.0,
                transform.translation.y
                    + (obstacle_height + star_width * 2.0)
                        * (if is_roof_obstacle { -1.0 } else { 1.0 }),
                6.0,
            );

            commands.spawn((
                SpriteBundle {
                    sprite: Sprite {
                        anchor: Anchor::Center,
                        ..default()
                    },
                    texture: texture,
                    transform: transform,
                    ..default()
                },
                Star {
                    radius: star_width,
                    points: star_value,
                },
            ));
        }
    }
}

fn despawn_offscreen(
    mut commands: Commands,
    window: Query<&Window, With<PrimaryWindow>>,
    player: Query<&Transform, With<Player>>,
    obstacles: Query<(Entity, &Transform), Or<(With<Obstacle>, With<Star>)>>,
) {
    let window = window.single();

    let Ok(player) = player.get_single() else {
        return;
    };

    for (entity, transform) in &obstacles {
        if transform.translation.x < player.translation.x - window.width() {
            commands.entity(entity).despawn();
        }
    }
}
