use std::time::Duration;

use bevy::prelude::*;

use crate::{
    asset_loader::AssetLoadState::AssetsLoaded,
    obstacle::{Obstacle, Star},
    system_set::{DespawnSet, MovementSet},
};

const PLANE_SPEED: f32 = 200.0;
const TAP_ACCELERATION: f32 = 50000.0;
const GRAVITY_ACCERATION_RATE: f32 = -1000.0;
// const GRAVITY_ACCERATION_RATE: f32 = 0.0;
const MAX_VELOCITY: f32 = 1000.0;
const MAX_TILT_DEGREES: f32 = 45.0;

pub struct PlayerPlugin;

impl Plugin for PlayerPlugin {
    fn build(&self, app: &mut App) {
        app.add_systems(OnEnter(AssetsLoaded), setup).add_systems(
            Update,
            (
                move_plane.in_set(MovementSet::PlayerMovement),
                animate_plane,
                obstacle_collider.before(DespawnSet::DespawnCollectedStars),
                collect_stars.in_set(DespawnSet::DespawnCollectedStars),
            ),
        );
    }
}

#[derive(Component)]
pub struct Player {
    acceleration: f32,
    velocity: f32,
    animation_index: usize,
    animation_timer: Timer,
    textures: Vec<Handle<Image>>,
}

#[derive(Component)]
struct PlayerCollider {
    radius: u32,
}

#[derive(Component)]
pub struct Score {
    points: u32,
}

impl Player {
    fn new(fps: u8, textures: Vec<Handle<Image>>) -> Self {
        Self {
            acceleration: 0.0,
            velocity: 0.0,
            animation_index: 0,
            animation_timer: Timer::new(
                Duration::from_secs_f32(1.0 / (fps as f32)),
                TimerMode::Repeating,
            ),
            textures: textures,
        }
    }
}

fn setup(mut commands: Commands, asset_server: Res<AssetServer>, assets: Res<Assets<Image>>) {
    let texture1: Handle<Image> = asset_server.load("textures/Planes/planeGreen1.png");
    let texture2: Handle<Image> = asset_server.load("textures/Planes/planeGreen2.png");
    let texture3: Handle<Image> = asset_server.load("textures/Planes/planeGreen3.png");

    let player_radius = assets.get(&texture1).unwrap().width() / 2;

    commands.spawn((
        SpriteBundle {
            texture: texture1.clone(),
            ..default()
        },
        Player::new(20, vec![texture1, texture2.clone(), texture3, texture2]),
        PlayerCollider {
            // Add a little wiggle room for collision detection
            radius: (player_radius as f32 * 0.8) as u32,
        },
        Score { points: 0 },
    ));
}

fn animate_plane(mut player: Query<(&mut Player, &mut Handle<Image>)>, time: Res<Time>) {
    let Ok((mut player, mut handle)) = player.get_single_mut() else {
        return;
    };

    player.animation_timer.tick(time.delta());

    if player.animation_timer.just_finished() {
        player.animation_index += 1;
        if player.animation_index == player.textures.len() {
            player.animation_index = 0;
        }
        *handle = player.textures[player.animation_index].clone();
    }
}

fn move_plane(
    mut player: Query<(&mut Player, &mut Transform)>,
    keys: Res<ButtonInput<KeyCode>>,
    time: Res<Time>,
) {
    let Ok((mut player, mut transform)) = player.get_single_mut() else {
        return;
    };

    transform.translation.x += PLANE_SPEED * time.delta_seconds();

    if keys.just_pressed(KeyCode::Space) {
        player.acceleration = TAP_ACCELERATION;
    } else {
        player.acceleration = GRAVITY_ACCERATION_RATE;
    }

    player.velocity += player.acceleration * time.delta_seconds();
    if player.velocity > MAX_VELOCITY {
        player.velocity = MAX_VELOCITY;
    }
    if player.velocity < -MAX_VELOCITY {
        player.velocity = -MAX_VELOCITY;
    }

    // println!("a {}, v {}", player.acceleration, player.velocity);

    transform.translation.y += player.velocity * time.delta_seconds();

    let degrees: f32 = player.velocity / MAX_VELOCITY * MAX_TILT_DEGREES;
    transform.rotation = Quat::from_rotation_z(degrees.to_radians());
}

fn obstacle_collider(player: Query<(&Transform, &PlayerCollider)>, obstacles: Query<&Obstacle>) {
    let Ok((player_transform, player_collider)) = player.get_single() else {
        return;
    };

    let player_position = Vec2::new(
        player_transform.translation.x,
        player_transform.translation.y,
    );
    let player_radius = player_collider.radius as f32;

    let mut collision_detected = false;

    for obstacle in &obstacles {
        if circle_collides_with_line_segment(
            obstacle.p1,
            obstacle.p2,
            player_position,
            player_radius,
        ) || circle_collides_with_line_segment(
            obstacle.p1,
            obstacle.p3,
            player_position,
            player_radius,
        ) || circle_collides_with_line_segment(
            obstacle.p2,
            obstacle.p3,
            player_position,
            player_radius,
        ) {
            collision_detected = true;
            break;
        }
    }

    if collision_detected {
        println!("Obstacle Collision detected");
    } else {
        // println!("No Collision detected");
    }
}

fn circle_collides_with_line_segment(a: Vec2, b: Vec2, c: Vec2, r: f32) -> bool {
    // Based on https://stackoverflow.com/a/1079478

    let ac = c - a;
    let ab = b - a;

    let d = ac.project_onto(ab) + a;

    let ad = d - a;

    let k = if ab.x.abs() > ab.y.abs() {
        ad.x / ab.x
    } else {
        ad.y / ab.y
    };

    let len_squared = if k <= 0.0 {
        (c - a).length_squared()
    } else if k >= 1.0 {
        (c - b).length_squared()
    } else {
        (c - d).length_squared()
    };

    return r * r > len_squared;
}

fn collect_stars(
    mut commands: Commands,
    mut player: Query<(&Transform, &PlayerCollider, &mut Score)>,
    stars: Query<(Entity, &Transform, &Star)>,
) {
    let Ok((player_transform, player_collider, mut score)) = player.get_single_mut() else {
        return;
    };

    let player_position = Vec2::new(
        player_transform.translation.x,
        player_transform.translation.y,
    );
    let player_radius = player_collider.radius as f32;

    for (entity, star_transform, star) in &stars {
        let star_position = Vec2::new(star_transform.translation.x, star_transform.translation.y);
        let star_radius = star.radius as f32;
        if circle_collides_with_circle(player_position, player_radius, star_position, star_radius) {
            score.points += star.points;
            println!("Player points: {}", score.points);
            commands.entity(entity).despawn();
        }
    }
}

fn circle_collides_with_circle(c1: Vec2, r1: f32, c2: Vec2, r2: f32) -> bool {
    return r1 * r1 + r2 * r2 > (c1 - c2).length_squared();
}
